import json
import unittest
from enum import Enum

from oldaplib.src.helpers.observable_set import ObservableSet
from oldaplib.src.helpers.serializer import serializer
from oldaplib.src.xsd.iri import Iri


class TestObservableSet(unittest.TestCase):

    _notified: Iri

    @classmethod
    def setUpClass(cls):
        cls._notified = ''

    def notifier_test(self, data: Iri) -> None:
        self._notified = data

    def test_constructor(self):
        obs = ObservableSet()
        self.assertEqual(obs, set())
        self.assertEqual(len(obs), 0)

        data = {'a', 'b', 'c'}
        obs = ObservableSet(data)
        self.assertEqual(obs, data)

        obs2 = ObservableSet(obs)
        self.assertEqual(obs2, data)

        obs = ObservableSet({'a', 'b', 'c'}, self.notifier_test, Iri('gaga:gaga'))
        self.assertEqual(obs, {'a', 'b', 'c'})

    def test_conversions(self):
        obs = ObservableSet({'a', 'b', 'c'}, self.notifier_test, Iri('gaga:gaga'))
        s = str(obs).strip('{}')
        s = s.split(", ")
        self.assertTrue("'a'" in s)
        self.assertTrue("'b'" in s)
        self.assertTrue("'c'" in s)

        obs = ObservableSet({'a', 'b', 'c'}, self.notifier_test, Iri('gaga:gaga'))
        s = repr(obs)
        s = s.split(", ")
        self.assertTrue("'a'" in s)
        self.assertTrue("'b'" in s)
        self.assertTrue("'c'" in s)

        obs = ObservableSet({'a', 'b', 'c'}, self.notifier_test, Iri('gaga:gaga'))
        s = obs.toRdf
        s = s.split(", ")
        self.assertTrue("a" in s)
        self.assertTrue("b" in s)
        self.assertTrue("c" in s)

    def test_logicals_or(self):
        obs = ObservableSet({'a', 'b', 'c'}, self.notifier_test, Iri('gaga:gaga'))
        res = obs | ObservableSet({'d', 'e', 'f'})
        self.assertEqual(res, {'a', 'b', 'c', 'd', 'e', 'f'})

        res = obs | {'d', 'e', 'f'}
        self.assertEqual(res, {'a', 'b', 'c', 'd', 'e', 'f'})

        self._notified = ''
        obs |= {'x', 'y'}
        self.assertEqual(obs, {'a', 'b', 'c', 'x', 'y'})
        self.assertEqual(self._notified, Iri('gaga:gaga'))

    def test_logicals_and(self):
        obs = ObservableSet({'a', 'b', 'c'}, self.notifier_test, Iri('gaga:gaga'))
        res = obs & ObservableSet({'b', 'c', 'd'})
        self.assertEqual(res, {'b', 'c'})

        res = obs & {'b', 'c', 'd'}
        self.assertEqual(res, {'b', 'c'})

        self._notified = ''
        obs &= {'b', 'c', 'd'}
        self.assertEqual(obs, {'b', 'c'})
        self.assertEqual(self._notified, Iri('gaga:gaga'))

    def test_logical_sub(self):
        obs = ObservableSet({'a', 'b', 'c', 'd'}, self.notifier_test, Iri('gaga:gaga'))
        res = obs - ObservableSet({'b', 'd'})
        self.assertEqual(res, {'a', 'c'})

        obs = ObservableSet({'a', 'b', 'c', 'd'}, self.notifier_test, Iri('gaga:gaga'))
        res = obs - {'b', 'd'}
        self.assertEqual(res, {'a', 'c'})

        self._notified = ''
        obs -= {'b', 'd'}
        self.assertEqual(obs, {'a', 'c'})
        self.assertEqual(self._notified, Iri('gaga:gaga'))

    def test_add(self):
        obs = ObservableSet({'a', 'b', 'c', 'd'}, self.notifier_test, Iri('gaga:gaga'))
        self._notified = ''
        obs.add('Z')
        self.assertEqual(obs, {'a', 'b', 'c', 'd', 'Z'})
        self.assertEqual(self._notified, Iri('gaga:gaga'))

    def test_remove(self):
        obs = ObservableSet({'a', 'b', 'c', 'd'}, self.notifier_test, Iri('gaga:gaga'))
        self._notified = ''
        obs.remove('c')
        self.assertEqual(obs, {'a', 'b', 'd'})
        self.assertEqual(self._notified, Iri('gaga:gaga'))

        obs = ObservableSet({'a', 'b', 'c', 'd'}, self.notifier_test, Iri('gaga:gaga'))
        with self.assertRaises(KeyError) as err:
            obs.remove('Z')

    def test_discard(self):
        obs = ObservableSet({'a', 'b', 'c', 'd'}, self.notifier_test, Iri('gaga:gaga'))
        self._notified = ''
        obs.discard('c')
        self.assertEqual(obs, {'a', 'b', 'd'})
        self.assertEqual(self._notified, 'gaga:gaga')

        obs = ObservableSet({'a', 'b', 'c', 'd'}, self.notifier_test, Iri('gaga:gaga'))
        self._notified = ''
        obs.discard('Z')
        self.assertEqual(obs, {'a', 'b', 'c', 'd'})
        self.assertEqual(self._notified, Iri('gaga:gaga'))

    def test_pop(self):
        obs = ObservableSet({'a', 'b', 'c', 'd'}, self.notifier_test, Iri('gaga:gaga'))
        self._notified = ''
        len1 = len(obs)
        obs.pop()
        len2 = len(obs)
        self.assertEqual(len1, len2 + 1)
        self.assertEqual(self._notified, Iri('gaga:gaga'))

        obs = ObservableSet()
        self._notified = ''
        len1 = len(obs)
        self.assertEqual(len1, 0)
        with self.assertRaises(KeyError):
            obs.pop()
        len2 = len(obs)
        self.assertEqual(len2, 0)
        self.assertEqual(self._notified, '')

    def test_clear(self):
        obs = ObservableSet({'a', 'b', 'c', 'd'}, self.notifier_test, Iri('gaga:gaga'))
        self._notified = ''
        obs.clear()
        self.assertEqual(obs, set())
        self.assertEqual(len(obs), 0)
        self.assertEqual(self._notified, Iri('gaga:gaga'))

    def test_copy(self):
        obs = ObservableSet({'a', 'b', 'c', 'd'}, self.notifier_test, Iri('gaga:gaga'))
        obs2 = obs.copy()
        obs.clear()
        self.assertEqual(obs, set())
        self.assertEqual(obs2, {'a', 'b', 'c', 'd'})

    def test_to_rdf(self):
        obs = ObservableSet({'a', 'b', 'c', 'd'}, self.notifier_test, Iri('gaga:gaga'))
        s = obs.toRdf
        s = s.split(', ')
        self.assertTrue('a' in s)
        self.assertTrue('b' in s)
        self.assertTrue('c' in s)
        self.assertTrue('d' in s)

    def test_json(self):
        obs = ObservableSet({'a', 'b', 'c', 'd'}, self.notifier_test, Iri('gaga:gaga'))
        jsonstr = json.dumps(obs, default=serializer.encoder_default)
        obs2 = json.loads(jsonstr, object_hook=serializer.decoder_hook)
        self.assertEqual(obs2, {'a', 'b', 'c', 'd'})

    def test_type_conversion(self):
        obs = ObservableSet({'a', 'b', 'c', 'd'}, self.notifier_test, Iri('gaga:gaga'))
        self.assertTrue(isinstance(obs.asSet(), set))


if __name__ == '__main__':
    unittest.main()



import json
from time import sleep

import bcrypt
import jwt
import requests

from jwt import InvalidTokenError
from pystrict import strict
from typing import Dict, Optional, Any
from datetime import datetime, timedelta
from rdflib.plugins.stores.sparqlstore import SPARQLUpdateStore
from pathlib import Path

from oldaplib.src.enums.adminpermissions import AdminPermission
from oldaplib.src.userdataclass import UserData
from oldaplib.src.xsd.xsd_qname import Xsd_QName
from oldaplib.src.xsd.xsd_ncname import Xsd_NCName
from oldaplib.src.helpers.oldaperror import OldapError, OldapErrorNoPermission
from oldaplib.src.helpers.context import Context, DEFAULT_CONTEXT
from oldaplib.src.helpers.query_processor import QueryProcessor
from oldaplib.src.helpers.serializer import serializer
from oldaplib.src.iconnection import IConnection
from oldaplib.src.enums.sparql_result_format import SparqlResultFormat
from oldaplib.src.xsd.xsd_string import Xsd_string

#
# For bootstrapping the whole tripel store, the following SPARQL has to be executed within the GraphDB
# SPARQL-console:
#
"""
PREFIX oldap: <http://oldap.org/base#>
PREFIX dcterms: <http://purl.org/dc/terms/>
PREFIX schema: <http://schema.org/>

INSERT DATA {
	GRAPH oldap:admin {
		<https://orcid.org/0000-0003-1681-4036> a oldap:User ;
        	oldap:userId "rosenth"^^xsd:NCName ;
		    dcterms:creator <https://orcid.org/0000-0003-1681-4036> ;
		    dcterms:created "2023-11-04T12:00:00+00:00"^^xsd:dateTime ;
		    dcterms:contributor <https://orcid.org/0000-0003-1681-4036> ;
		    dcterms:modified "2023-11-04T12:00:00+00:00"^^xsd:dateTime ;
        	schema:familyName "Rosenthaler"^^xsd:string ;
        	schema:givenName "Lukas"^^xsd:string ;
        	oldap:credentials "$2b$12$N00UMBBJG9XfPV6R5NxulOTKi0qRBpypTFe82dKwSdTFrWZS7nat2"^^xsd:string ;
        	oldap:isActive "true"^^xsd:boolean .
	}
}

"""
#
# Then, executing the __main__ of the file "connection.py" will initialize the triple store with all the data
# needed to run the tests
#
jwt_format = {
    "userId": "https://orcid.org/0000-0003-1681-4036",
    "exp": "2023-11-04T12:00:00+00:00",
    "iat": datetime.now().isoformat(),
    "iss": "http://oldap.org"
}
token = "eyJhbGciOiJIUzI1NiIsInR5cCI6IkpXVCJ9.eyJ1c2VySWQiOiIkMmIkMTIkaldDSloucWRYRTlNU0NQZFVjMHk0Ljlzd1dZSmNnTFpuMGVQdFJUdS83VThxSC9PWFhrQjIiLCJleHAiOiIyMDI0LTExLTA0VDEyOjAwOjAwKzAwOjAwIiwiaWF0IjoiMjAyNC0wMS0xOVQyMzo0MTozMS45NTI5MTkiLCJpc3MiOiJodHRwOi8vb2xkYXAub3JnIn0.Vsc2qamfyeTW6Xz5l2Wca-mFnA5PcLuOoWPVEo__4Z4"


class Connection(IConnection):
    """
    Class that implements the connection to an external triple store for oldap.

    The connection to a SPARQL endpoint requires the following information:

    * _server_: URL of the server (inlcuding port number)
    * _repo_: Name of the repository to connect to
    * _context_name_: Name of the context (see ~helper.Context)

    The class implements the following methods:

    * Getter-methods:
        - _server_: returns the server string
        - _repo_: returns the repository name
        - _context_name_: returns the context name
    * Setter methods:
        - Any setting of the _server_, _repo_ or _context_name_ - variables raises an OldapError-exception
    * Further methods
        - _Constructor(server,repo,contextname)_: requires _server_ and _repo_string, _context_name defaults to "DEFAULT"
        - _clear_graph_(graph_name: QName)_: Deletes the given graph (must be given as QName)
        - _clear_repo()_ Deletes all data in the repository given by the Connection instance
        - _upload_turtle(filename: str, graphname:str)_: Loads the data in the given file (must be turtle or trig
          format) into the given repo and graph. If graphname is not given, the data will either be loaded into the
          default graph of the repository or into the graph given in the trig file.
          _Note_: The method returns before the triple store has digested all the data! It may not immediately
          available after this method returns!
        - _query(query: str, format: SparqlResultFormat)_: Sends a SPARQL query to the triple store and returns the
          result in the given format. If no format is given, JSON will be used as default format.
        - _update_query(query: str)_: Send a SPARQL update query to the SPARQL endpoint. The method return either
          {'status': 'OK'} or {'status': 'ERROR', 'message': 'error-text'}
        - _rdflib_query(query: str, bindings: Optional[Mapping[str, Identifier]])_: Send a SPAQRL query using rdflib
          to the SPARQL endpoint. The variable _bindings_ allows to set query parameters to given values.

    """
    _server: str
    _repo: str
    _userId: str
    _store: SPARQLUpdateStore
    _query_url: str
    _update_url: str
    _transaction_url: Optional[str]
    jwtkey: str = "You have to change this!!! +D&RWG+"
    _switcher = {
        SparqlResultFormat.XML: lambda a: a.text,
        SparqlResultFormat.JSON: lambda a: a.json(),
        SparqlResultFormat.TURTLE: lambda a: a.text,
        SparqlResultFormat.N3: lambda a: a.text,
        SparqlResultFormat.NQUADS: lambda a: a.text,
        SparqlResultFormat.JSONLD: lambda a: a.json(),
        SparqlResultFormat.TRIX: lambda a: a.text,
        SparqlResultFormat.TRIG: lambda a: a.text,
        SparqlResultFormat.TEXT: lambda a: a.text
    }

    def __init__(self, *,
                 server: str,
                 repo: str,
                 userId: Optional[str | Xsd_NCName] = None,
                 credentials: Optional[str | Xsd_string] = None,
                 token: Optional[str] = None,
                 context_name: Optional[str] = DEFAULT_CONTEXT) -> None:
        """
        Constructor that establishes the connection parameters.

        :param server: URL of the server (including port information if necessary)
        :param repo: Name of the triple store repository on the server
        :param context_name: A name of the Context to be used (see ~Context). If no such context exists,
            a new context of this name is created
        """
        super().__init__(context_name=context_name)
        self._server = server
        self._repo = repo
        self._query_url = f'{self._server}/repositories/{self._repo}'
        self._update_url = f'{self._server}/repositories/{self._repo}/statements'
        self._store = SPARQLUpdateStore(self._query_url, self._update_url)
        context = Context(name=context_name)
        if token is not None:
            try:
                payload = jwt.decode(jwt=token, key=Connection.jwtkey, algorithms="HS256")
            except InvalidTokenError:
                raise OldapError("Wrong credentials")
            self._userdata = json.loads(payload['userdata'], object_hook=serializer.decoder_hook)
            self._token = token
            return
        if userId is None and credentials is None:
            userId = Xsd_NCName("unknown", validate=False)
        if not isinstance(userId, Xsd_NCName):
            userId = Xsd_NCName(userId)
        if userId is None or credentials is None:
            raise OldapError("Wrong credentials")
        sparql = UserData.sparql_query(context=context, userId=userId)
        headers = {
            "Content-Type": "application/x-www-form-urlencoded; charset=UTF-8",
            "Accept": "application/x-sparqlstar-results+json, application/sparql-results+json;q=0.9, */*;q=0.8",
        }
        data = {
            'query': sparql,
        }
        res = requests.post(url=self._query_url, headers=headers, data=data)
        if res.status_code == 200:
            jsonobj = res.json()
        else:
            raise OldapError(res.status_code, res.text)
        res = QueryProcessor(context=context, query_result=jsonobj)

        self._userdata = UserData.from_query(res)
        if not self._userdata.isActive:
            raise OldapError("Wrong credentials")  # On purpose, we are not providing too much information why the login failed
        if userId != "unknown":
            hashed = str(self._userdata.credentials).encode('utf-8')
            if not bcrypt.checkpw(credentials.encode('utf-8'), hashed):
                raise OldapError("Wrong credentials")  # On purpose, we are not providing too much information why the login failed

        expiration = datetime.now() + timedelta(days=1)
        payload = {
            "userdata": json.dumps(self._userdata, default=serializer.encoder_default),
            "exp": expiration.timestamp(),
            "iat": int(datetime.now().timestamp()),
            "iss": "http://oldap.org"
        }
        self._token = jwt.encode(
            payload=payload,
            key=Connection.jwtkey,
            algorithm="HS256")
        #
        # Get projects and add to Context
        #
        sparql = context.sparql_context
        sparql += """
        SELECT ?sname ?ns
        FROM oldap:onto
        FROM shared:onto
        FROM NAMED oldap:admin
        WHERE {
            GRAPH oldap:admin {
                ?proj a oldap:Project .
                ?proj oldap:projectShortName ?sname .
                ?proj oldap:namespaceIri ?ns .
            }
        }
        """
        headers = {
            "Content-Type": "application/x-www-form-urlencoded; charset=UTF-8",
            "Accept": "application/x-sparqlstar-results+json, application/sparql-results+json;q=0.9, */*;q=0.8",
        }
        data = {
            'query': sparql,
        }
        res = requests.post(url=self._query_url, headers=headers, data=data)
        if res.status_code == 200:
            jsonobj = res.json()
        else:
            raise OldapError(res.status_code, res.text)
        res = QueryProcessor(context=context, query_result=jsonobj)
        for r in res:
            context[r['sname']] = r['ns']

    @property
    def server(self) -> str:
        """Getter for server string"""
        return self._server

    @property
    def repo(self) -> str:
        """Getter for repository name"""
        return self._repo

    def clear_graph(self, graph_iri: Xsd_QName) -> None:
        """
        This method clears (deletes) the given RDF graph. May raise an OldapError.

        :param graph_iri: RDF graph name as QName. The prefix must be defined in the context!
        :return: None
        """
        if not self._userdata:
            raise OldapErrorNoPermission("No permission")
        actor = self._userdata
        sysperms = actor.inProject.get(Xsd_QName('oldap:SystemProject'))
        is_root: bool = False
        if sysperms and AdminPermission.ADMIN_OLDAP in sysperms:
            is_root = True
        if not is_root:
            raise OldapErrorNoPermission("No permission")

        context = Context(name=self._context_name)
        headers = {
            "Content-Type": "application/sparql-update",
            "Accept": "application/json, text/plain, */*",
        }
        data = f"CLEAR GRAPH <{context.qname2iri(graph_iri)}>"
        req = requests.post(self._update_url,
                            headers=headers,
                            data=data)
        if not req.ok:
            raise OldapError(req.text)

    def clear_repo(self) -> None:
        """
        This method deletes the complete repository. Use with caution!!!

        :return: None
        """
        # if not self._userdata:
        #     raise OldapErrorNoPermission("No permission")
        # actor = self._userdata
        # sysperms = actor.inProject.get(QName('oldap:SystemProject'))
        # is_root: bool = False
        # if sysperms and AdminPermission.ADMIN_OLDAP in sysperms:
        #     is_root = True
        # if not is_root:
        #     raise OldapErrorNoPermission("No permission")
        headers = {
            "Accept": "application/json, text/plain, */*",
        }
        data = {"update": "CLEAR ALL"}
        req = requests.post(self._update_url,
                            headers=headers,
                            data=data)
        if not req.ok:
            raise OldapError(req.text)

    def upload_turtle(self, filename: str, graphname: Optional[str] = None) -> None:
        """
        Upload a turtle- or trig-file to the given repository. This method returns immediately after sending the
        command to upload the given file to the triplestore. The import process may take a while!

        :param filename: Name of the file to upload
        :param graphname: Optional name of the RDF-graph where the data should be imported in.
        :return: None
        """
        # if not self._userdata:
        #     raise OldapErrorNoPermission("No permission")
        # actor = self._userdata
        # sysperms = actor.inProject.get(QName('oldap:SystemProject'))
        # is_root: bool = False
        # if sysperms and AdminPermission.ADMIN_OLDAP in sysperms:
        #     is_root = True
        # if not is_root:
        #     raise OldapErrorNoPermission("No permission")

        with open(filename, encoding="utf-8") as f:
            content = f.read()
            ext = Path(filename).suffix
            mime = ""
            if ext == ".ttl":
                mime = "text/turtle"
            elif ext == ".trig":
                mime = "application/x-trig"

            ct = datetime.now()
            ts = ct.timestamp()
            data = {
                "name": f'Data from "{filename}"',
                "status": None,
                "message": "",
                "context": graphname,
                "replaceGraphs": [],
                "baseURI": None,
                "forceSerial": False,
                "type": "text",
                "format": mime,
                "data": content,
                "timestamp": ts,
                "parserSettings": {
                    "preserveBNodeIds": False,
                    "failOnUnknownDataTypes": False,
                    "verifyDataTypeValues": False,
                    "normalizeDataTypeValues": False,
                    "failOnUnknownLanguageTags": False,
                    "verifyLanguageTags": True,
                    "normalizeLanguageTags": True,
                    "stopOnError": True
                }
            }
            jsondata = json.dumps(data)
            headers = {
                "Accept": "application/json, text/plain, */*",
                "Content-Type": "application/json; charset=utf-8"
            }
            url = f"{self._server}/rest/repositories/{self._repo}/import/upload/text"
            req = requests.post(url,
                                headers=headers,
                                data=jsondata)
        if not req.ok:
            raise OldapError(req.text)

    def query(self, query: str, format: SparqlResultFormat = SparqlResultFormat.JSON) -> Any:
        """
        Send a SPARQL-query and return the result. The result may be nested dict (in case of JSON) or a text.

        :param query: SPARQL query as string
        :param format: The format desired (see ~SparqlResultFormat)
        :return: Query results or an error message (as text)
        """
        if not self._userdata:
            raise OldapError("No login")
        headers = {
            "Content-Type": "application/x-www-form-urlencoded; charset=UTF-8",
            "Accept": format.value,
        }
        data = {
            'query': query,
        }
        res = requests.post(url=self._query_url,
                            headers=headers,
                            data=data)
        if res.status_code == 200:
            return Connection._switcher[format](res)
        else:
            raise OldapError(res.text)

    def update_query(self, query: str) -> Dict[str,str]:
        """
        Send an SPARQL UPDATE query to the triple store
        :param query: SPARQL UPDATE query as string
        :return:
        """
        if not self._userdata:
            raise OldapError("No login")
        headers = {
            "Accept": "*/*"
        }
        url = f"{self._server}/repositories/{self._repo}/statements"
        res = requests.post(url, data={"update": query}, headers=headers)
        if not res.ok:
            raise OldapError(f'Update query failed. Reason: "{res.text}"')

    def transaction_start(self) -> None:
        if not self._userdata:
            raise OldapError("No login")
        headers = {
            "Accept": "*/*"
        }
        url = f"{self._server}/repositories/{self._repo}/transactions"
        res = requests.post(url, headers=headers)
        if res.headers.get('location') is None:
            raise OldapError('GraphDB start of transaction failed')
        self._transaction_url = res.headers['location']

    def transaction_query(self, query: str, result_format: SparqlResultFormat = SparqlResultFormat.JSON) -> Any:
        if not self._userdata:
            raise OldapError("No login")
        headers = {
            "Content-Type": "application/x-www-form-urlencoded; charset=UTF-8",
            "Accept": result_format.value
        }
        if self._transaction_url is None:
            raise OldapError("No GraphDB transaction started")
        res = requests.post(self._transaction_url, data={'action': 'QUERY', 'query': query}, headers=headers)
        if not res.ok:
            raise OldapError(f'GraphDB Transaction query failed. Reason: "{res.text}"')
        return Connection._switcher[result_format](res)

    def transaction_update(self, query: str) -> None:
        if not self._userdata:
            raise OldapError("No login")
        headers = {
            "Content-Type": "application/x-www-form-urlencoded; charset=UTF-8",
            "Accept": "*/*"
        }
        if self._transaction_url is None:
            raise OldapError("No GraphDB transaction started")
        res = requests.post(self._transaction_url, data={'action': 'UPDATE', 'update': query}, headers=headers)
        if not res.ok:
            raise OldapError(f'GraphDB Transaction update failed. Reason: "{res.text}"')

    def transaction_commit(self) -> None:
        if not self._userdata:
            raise OldapError("No login")
        headers = {
            "Accept": "*/*"
        }
        if self._transaction_url is None:
            raise OldapError("No GraphDB transaction started")
        res = requests.put(f'{self._transaction_url}?action=COMMIT', headers=headers)
        if not res.ok:
            raise OldapError(f'GraphDB transaction commit failed. Reason: "{res.text}"')
        self._transaction_url = None

    def transaction_abort(self) -> None:
        if not self._userdata:
            raise OldapError("No login")
        headers = {
            "Accept": "*/*"
        }
        if self._transaction_url is None:
            raise OldapError("No GraphDB transaction started")
        res = requests.delete(self._transaction_url, headers=headers)
        if not res.ok:
            raise OldapError(f'GraphDB transaction abort failed. Reason: "{res.text}"')
        self._transaction_url = None

    def in_transaction(self) -> bool:
        return self._transaction_url is not None



    # def rdflib_query(self, query: str,
    #                  bindings: Optional[Mapping[str, Identifier]] = None) -> Result:
    #     """
    #     Send a SPARQL query to a triple store using the Python rdflib interface.
    #
    #     :param query: SPARQL query string
    #     :param bindings: Bindings to variables
    #     :return: a RDFLib Result instance
    #     """
    #     if not self._userdata:
    #         raise OldapError("No login")
    #     return self._store.query(query, initBindings=bindings)


if __name__ == "__main__":
    con = Connection(server='http://localhost:7200',
                     userId="rosenth",
                     credentials="RioGrande",
                     repo="oldap",
                     context_name="DEFAULT")
    con.clear_repo()
    con.upload_turtle("../ontologies/oldap.trig")
    con.upload_turtle("../ontologies/shared.trig")
    con.upload_turtle("../ontologies/admin.trig")




import re
import threading
from datetime import datetime
from enum import Enum, EnumMeta
from functools import partial
from pprint import pprint
from typing import Type, Any, Self

from oldaplib.src.datamodel import DataModel
from oldaplib.src.enums.action import Action
from oldaplib.src.enums.datapermissions import DataPermission
from oldaplib.src.enums.haspropertyattr import HasPropertyAttr
from oldaplib.src.enums.adminpermissions import AdminPermission
from oldaplib.src.enums.propertyclassattr import PropClassAttr
from oldaplib.src.enums.xsd_datatypes import XsdDatatypes
from oldaplib.src.hasproperty import HasProperty
from oldaplib.src.helpers.attributechange import AttributeChange
from oldaplib.src.helpers.context import Context
from oldaplib.src.helpers.convert2datatype import convert2datatype
from oldaplib.src.helpers.langstring import LangString
from oldaplib.src.helpers.observable_set import ObservableSet
from oldaplib.src.helpers.oldaperror import OldapErrorNotFound, OldapErrorValue, OldapErrorInconsistency, \
    OldapErrorNoPermission, OldapError, OldapErrorUpdateFailed, OldapErrorInUse
from oldaplib.src.helpers.query_processor import QueryProcessor
from oldaplib.src.iconnection import IConnection
from oldaplib.src.project import Project
from oldaplib.src.propertyclass import PropertyClass
from oldaplib.src.resourceclass import ResourceClass
from oldaplib.src.xsd.iri import Iri
from oldaplib.src.xsd.xsd import Xsd
from oldaplib.src.xsd.xsd_anyuri import Xsd_anyURI
from oldaplib.src.xsd.xsd_base64binary import Xsd_base64Binary
from oldaplib.src.xsd.xsd_boolean import Xsd_boolean
from oldaplib.src.xsd.xsd_byte import Xsd_byte
from oldaplib.src.xsd.xsd_date import Xsd_date
from oldaplib.src.xsd.xsd_datetime import Xsd_dateTime
from oldaplib.src.xsd.xsd_datetimestamp import Xsd_dateTimeStamp
from oldaplib.src.xsd.xsd_decimal import Xsd_decimal
from oldaplib.src.xsd.xsd_double import Xsd_double
from oldaplib.src.xsd.xsd_duration import Xsd_duration
from oldaplib.src.xsd.xsd_float import Xsd_float
from oldaplib.src.xsd.xsd_gday import Xsd_gDay
from oldaplib.src.xsd.xsd_gmonth import Xsd_gMonth
from oldaplib.src.xsd.xsd_gmonthday import Xsd_gMonthDay
from oldaplib.src.xsd.xsd_gyear import Xsd_gYear
from oldaplib.src.xsd.xsd_gyearmonth import Xsd_gYearMonth
from oldaplib.src.xsd.xsd_hexbinary import Xsd_hexBinary
from oldaplib.src.xsd.xsd_id import Xsd_ID
from oldaplib.src.xsd.xsd_idref import Xsd_IDREF
from oldaplib.src.xsd.xsd_int import Xsd_int
from oldaplib.src.xsd.xsd_language import Xsd_language
from oldaplib.src.xsd.xsd_long import Xsd_long
from oldaplib.src.xsd.xsd_ncname import Xsd_NCName
from oldaplib.src.xsd.xsd_negativeinteger import Xsd_negativeInteger
from oldaplib.src.xsd.xsd_nmtoken import Xsd_NMTOKEN
from oldaplib.src.xsd.xsd_nonnegativeinteger import Xsd_nonNegativeInteger
from oldaplib.src.xsd.xsd_nonpositiveinteger import Xsd_nonPositiveInteger
from oldaplib.src.xsd.xsd_normalizedstring import Xsd_normalizedString
from oldaplib.src.xsd.xsd_positiveinteger import Xsd_positiveInteger
from oldaplib.src.xsd.xsd_qname import Xsd_QName
from oldaplib.src.xsd.xsd_short import Xsd_short
from oldaplib.src.xsd.xsd_string import Xsd_string
from oldaplib.src.xsd.xsd_time import Xsd_time
from oldaplib.src.xsd.xsd_token import Xsd_token
from oldaplib.src.xsd.xsd_unsignedbyte import Xsd_unsignedByte
from oldaplib.src.xsd.xsd_unsignedint import Xsd_unsignedInt
from oldaplib.src.xsd.xsd_unsignedlong import Xsd_unsignedLong
from oldaplib.src.xsd.xsd_unsignedshort import Xsd_unsignedShort

ValueType = LangString | ObservableSet


#@strict
class ResourceInstance:
    _iri: Iri
    _values: dict[Iri, LangString | ObservableSet]
    _graph: Xsd_NCName
    _changeset: dict[Iri, AttributeChange]

    #@staticmethod
    # def convert2datatype(value: Any, datatype: XsdDatatypes) -> Xsd | LangString:
    #     match datatype:
    #         case XsdDatatypes.string:
    #             return Xsd_string(value)
    #         case XsdDatatypes.langString:
    #             return LangString(value)
    #         case XsdDatatypes.boolean:
    #             return Xsd_boolean(value)
    #         case XsdDatatypes.decimal:
    #             return Xsd_decimal(value)
    #         case XsdDatatypes.float:
    #             return Xsd_float(value)
    #         case XsdDatatypes.double:
    #             return Xsd_double(value)
    #         case XsdDatatypes.duration:
    #             return Xsd_duration(value)
    #         case XsdDatatypes.dateTime:
    #             return Xsd_dateTime(value)
    #         case XsdDatatypes.dateTimeStamp:
    #             return Xsd_dateTimeStamp(value)
    #         case XsdDatatypes.time:
    #             return Xsd_time(value)
    #         case XsdDatatypes.date:
    #             return Xsd_date(value)
    #         case XsdDatatypes.gYearMonth:
    #             return Xsd_gYearMonth(value)
    #         case XsdDatatypes.gYear:
    #             return Xsd_gYear(value)
    #         case XsdDatatypes.gMonthDay:
    #             return Xsd_gMonthDay(value)
    #         case XsdDatatypes.gDay:
    #             return Xsd_gDay(value)
    #         case XsdDatatypes.gMonth:
    #             return Xsd_gMonth(value)
    #         case XsdDatatypes.hexBinary:
    #             return Xsd_hexBinary(value)
    #         case XsdDatatypes.base64Binary:
    #             return Xsd_base64Binary(value)
    #         case XsdDatatypes.anyURI:
    #             return Xsd_anyURI(value)
    #         case XsdDatatypes.QName:
    #             return Xsd_QName(value)
    #         case XsdDatatypes.normalizedString:
    #             return Xsd_normalizedString(value)
    #         case XsdDatatypes.token:
    #             return Xsd_token(value)
    #         case XsdDatatypes.language:
    #             return Xsd_language(value)
    #         case XsdDatatypes.NCName:
    #             return Xsd_NCName(value)
    #         case XsdDatatypes.NMTOKEN:
    #             return Xsd_NMTOKEN(value)
    #         case XsdDatatypes.ID:
    #             return Xsd_ID(value)
    #         case XsdDatatypes.IDREF:
    #             return Xsd_IDREF(value)
    #         case XsdDatatypes.integer:
    #             return Xsd_int(value)
    #         case XsdDatatypes.nonPositiveInteger:
    #             return Xsd_nonPositiveInteger(value)
    #         case XsdDatatypes.negativeInteger:
    #             return Xsd_negativeInteger(value)
    #         case XsdDatatypes.long:
    #             return Xsd_long(value)
    #         case XsdDatatypes.int:
    #             return Xsd_int(value)
    #         case XsdDatatypes.short:
    #             return Xsd_short(value)
    #         case XsdDatatypes.byte:
    #             return Xsd_byte(value)
    #         case XsdDatatypes.nonNegativeInteger:
    #             return Xsd_nonNegativeInteger(value)
    #         case XsdDatatypes.unsignedLong:
    #             return Xsd_unsignedLong(value)
    #         case XsdDatatypes.unsignedInt:
    #             return Xsd_unsignedInt(value)
    #         case XsdDatatypes.unsignedShort:
    #             return Xsd_unsignedShort(value)
    #         case XsdDatatypes.unsignedByte:
    #             return Xsd_unsignedByte(value)
    #         case XsdDatatypes.positiveInteger:
    #             return Xsd_positiveInteger(value)
    #         case None:
    #             return Iri(value)
    #         case _:
    #             raise OldapErrorValue(f'Invalid datatype "{datatype}" for value "{value}"')

    def __init__(self, *,
                 iri: Iri | None = None,
                 **kwargs):
        self._iri = iri or Iri()
        self._values = {}
        self._graph = self.project.projectShortName
        self._superclass_objs = {}
        self._changeset = {}

        def set_values(propclass: dict[Iri, HasProperty]):
            for prop_iri, hasprop in propclass.items():
                if kwargs.get(prop_iri.fragment):
                    value = kwargs[prop_iri.fragment]
                    if isinstance(value, (list, tuple, set, LangString)):  # we may have multiple values...
                        if hasprop.prop.datatype == XsdDatatypes.langString:
                            self._values[prop_iri] = LangString(value,
                                                                notifier=self.notifier, notify_data=prop_iri)
                        else:
                            self._values[prop_iri] = ObservableSet({convert2datatype(x, hasprop.prop.datatype) for x in value},
                                                                   notifier=self.notifier, notify_data=prop_iri)
                    else:
                        try:
                            self._values[prop_iri] = ObservableSet({convert2datatype(value, hasprop.prop.datatype)})
                        except TypeError as err:
                            self._values[prop_iri] = convert2datatype(value, hasprop.prop.datatype)

            for prop_iri, hasprop in propclass.items():
                #
                # Validate
                #
                if hasprop.get(HasPropertyAttr.MIN_COUNT):  # testing for MIN_COUNT conformance
                    if hasprop[HasPropertyAttr.MIN_COUNT] > 0 and not self._values.get(prop_iri):
                        raise OldapErrorValue(f'{self.name}: Property {prop_iri} with MIN_COUNT={hasprop[HasPropertyAttr.MIN_COUNT]} is missing')
                    elif isinstance(self._values[prop_iri], ObservableSet) and len(self._values[prop_iri]) < 1:
                        raise OldapErrorValue(f'{self.name}: Property {prop_iri} with MIN_COUNT={hasprop[HasPropertyAttr.MIN_COUNT]} is missing')
                if hasprop.get(HasPropertyAttr.MAX_COUNT):  # testing for MAX_COUNT conformance
                    if isinstance(self._values.get(prop_iri), ObservableSet) and len(self._values[prop_iri]) > hasprop[HasPropertyAttr.MAX_COUNT]:
                        raise OldapErrorValue(f'{self.name}: Property {prop_iri} with MAX_COUNT={hasprop[HasPropertyAttr.MIN_COUNT]} has to many values (n={len(self._values[prop_iri.fragment])})')
                if self._values.get(prop_iri):
                    if isinstance(self._values[prop_iri], LangString):
                        self.validate_value(self._values[prop_iri], hasprop.prop)
                    else:
                        if isinstance(self._values[prop_iri], ObservableSet):
                            for val in self._values[prop_iri]:
                                self.validate_value(val, hasprop.prop)
                        else:
                            self.validate_value(self._values[prop_iri], hasprop.prop)

        def process_superclasses(superclass: dict[Iri, ResourceClass]):
           for sc_iri, sc in superclass.items():
                if sc.superclass:
                    process_superclasses(sc.superclass)
                if sc.owl_class_iri == Iri("oldap:Thing", validate=False):
                    timestamp = Xsd_dateTimeStamp()
                    if not self._values.get('oldap:createdBy'):
                        self._values[Iri('oldap:createdBy', validate=False)] = ObservableSet({self._con.userIri})
                    if not self._values.get('oldap:creationDate'):
                        self._values[Iri('oldap:creationDate', validate=False)] = ObservableSet({timestamp})
                    if not self._values.get('oldap:lastModifiedBy'):
                        self._values[Iri('oldap:lastModifiedBy', validate=False)] = ObservableSet({self._con.userIri})
                    if not self._values.get('oldap:lastModificationDate'):
                        self._values[Iri('oldap:lastModificationDate', validate=False)] = ObservableSet({timestamp})
                set_values(sc.properties)
                for iri, prop in sc.properties.items():
                    self.properties[iri] = prop


        if self.superclass:
            process_superclasses(self.superclass)

        set_values(self.properties)

        for propname in self.properties.keys():
            setattr(ResourceInstance, propname.fragment, property(
                partial(ResourceInstance.__get_value, prefix=propname.prefix, fragment=propname.fragment),
                partial(ResourceInstance.__set_value, prefix=propname.prefix, fragment=propname.fragment),
                partial(ResourceInstance.__del_value, prefix=propname.prefix, fragment=propname.fragment)))
        self.clear_changeset()

    def validate_value(self, values: ValueType, property: PropertyClass):
        if property.get(PropClassAttr.LANGUAGE_IN):  # testing for LANGUAGE_IN conformance
            if not isinstance(values, LangString):
                raise OldapErrorInconsistency(f'Property {property.property_class_iri} with LANGUAGE_IN requires datatype rdf:langstring, got {type(values).__name__}.')
            for lang, dummy in values.items():
                if not lang in property[PropClassAttr.LANGUAGE_IN]:
                    raise OldapErrorValue(f'Property {property.property_class_iri} with LANGUAGE_IN={property[PropClassAttr.LANGUAGE_IN]} has invalid language "{lang.value}"')
        if property.get(PropClassAttr.UNIQUE_LANG):
            return  # TODO: LangString does not yet allow multiple entries of the same language...
        if property.get(PropClassAttr.IN):
            for val in values:
                if not val in property[PropClassAttr.IN]:
                    raise OldapErrorValue(f'Property {property} with IN={property[PropClassAttr.IN]} has invalid value "{val}"')
        if property.get(PropClassAttr.MIN_LENGTH):
            for val in values:
                l = 0
                try:
                    l = len(val)
                except TypeError:
                    raise OldapErrorInconsistency(f'Property {property} with MIN_LENGTH={property[PropClassAttr.MIN_LENGTH]} has no length.')
                if l < property[PropClassAttr.MIN_LENGTH]:
                    raise OldapErrorInconsistency(f'Property {property} with MIN_LENGTH={property[PropClassAttr.MIN_LENGTH]} has length "{len(val)}".')
        if property.get(PropClassAttr.MAX_LENGTH):
            for val in values:
                l = 0
                try:
                    l = len(val)
                except TypeError:
                    raise OldapErrorInconsistency(f'Property {property} with MAX_LENGTH={property[PropClassAttr.MAX_LENGTH]} has no length.')
                if l > property[PropClassAttr.MAX_LENGTH]:
                    raise OldapErrorInconsistency(f'Property {property} with MIN_LENGTH={property[PropClassAttr.MAX_LENGTH]} has length "{len(val)}".')
        if property.get(PropClassAttr.PATTERN):
            for val in values:
                if not re.fullmatch(str(property[PropClassAttr.PATTERN]), str(val)):
                    raise OldapErrorInconsistency(f'Property {property} with PATTERN={property[PropClassAttr.PATTERN]} does not conform ({val}).')
        if property.get(PropClassAttr.MIN_EXCLUSIVE):
            for val in values:
                v: bool | None = None
                try:
                    v = val > property[PropClassAttr.MIN_EXCLUSIVE]
                except TypeError:
                    raise OldapErrorInconsistency(f'Property {property} with MIN_EXCLUSIVE={property[PropClassAttr.MIN_EXCLUSIVE]} cannot be compared to "{val}".')
                if not v:
                    raise OldapErrorInconsistency(f'Property {property} with MIN_EXCLUSIVE={property[PropClassAttr.MIN_EXCLUSIVE]} has invalid "{val}".')
        if property.get(PropClassAttr.MIN_INCLUSIVE):
            for val in values:
                v: bool | None = None
                try:
                    v = val >= property[PropClassAttr.MIN_INCLUSIVE]
                except TypeError:
                    raise OldapErrorInconsistency(f'Property {property} with MIN_EXCLUSIVE={property[PropClassAttr.MIN_INCLUSIVE]} cannot be compared to "{value}".')
                if not v:
                    raise OldapErrorInconsistency(f'Property {property} with MIN_EXCLUSIVE={property[PropClassAttr.MIN_INCLUSIVE]} has invalid "{value}".')
        if property.get(PropClassAttr.MAX_EXCLUSIVE):
            for val in values:
                v: bool | None = None
                try:
                    v = val < property[PropClassAttr.MAX_EXCLUSIVE]
                except TypeError:
                    raise OldapErrorInconsistency(f'Property {property} with MAX_EXCLUSIVE={property[PropClassAttr.MAX_EXCLUSIVE]} cannot be compared to "{value}".')
                if not v:
                    raise OldapErrorInconsistency(f'Property {property} with MAX_EXCLUSIVE={property[PropClassAttr.MAX_EXCLUSIVE]} has invalid "{value}".')
        if property.get(PropClassAttr.MAX_INCLUSIVE):
            for val in values:
                v: bool | None = None
                try:
                    v = val <= property[PropClassAttr.MAX_INCLUSIVE]
                except TypeError:
                    raise OldapErrorInconsistency(f'Property {property} with MAX_INCLUSIVE={property[PropClassAttr.MAX_INCLUSIVE]} cannot be compared to "{value}".')
                if not v:
                    raise OldapErrorInconsistency(f'Property {property} with MAX_INCLUSIVE={property[PropClassAttr.MAX_INCLUSIVE]} has invalid "{value}".')
        if property.get(PropClassAttr.LESS_THAN):
            other_values = self._values.get(property[PropClassAttr.LESS_THAN])
            if other_values is not None:
                b: bool | None = None
                try:
                    min_other_value = min(other_values)
                    max_value = max(values)
                    b = max_value < min_other_value
                except TypeError:
                    raise OldapErrorInconsistency(
                        f'Property {property} with LESS_THAN={property[PropClassAttr.LESS_THAN]} cannot be compared "{max_value} / {min_other_value}".')
                if not b:
                    raise OldapErrorInconsistency(
                        f'Property {property} with LESS_THAN={property[PropClassAttr.LESS_THAN]} has invalid value: "{max_value}" NOT LESS_THAN "{min_other_value}".')
        if property.get(PropClassAttr.LESS_THAN_OR_EQUALS):
            other_values = self._values.get(property[PropClassAttr.LESS_THAN])
            if other_values is not None:
                b: bool | None = None
                try:
                    min_other_value = min(other_values)
                    max_value = max(values)
                    b = max_value <= min_other_value
                except TypeError:
                    raise OldapErrorInconsistency(
                        f'Property {property} with LESS_THAN={property[PropClassAttr.LESS_THAN_OR_EQUALS]} cannot be compared "{max_value} / {min_other_value}".')
                if not b:
                    raise OldapErrorInconsistency(
                        f'Property {property} with LESS_THAN={property[PropClassAttr.LESS_THAN_OR_EQUALS]} has invalid value: "{max_value}" NOT LESS_THAN "{min_other_value}".')

    def notifier(self, prop_iri: Iri):
        hasprop = self.properties[prop_iri]
        self.validate_value(self._values[prop_iri], hasprop)

        if hasprop.get(HasPropertyAttr.MIN_COUNT):  # testing for MIN_COUNT conformance
            n = len(self._values[prop_iri])
            if n < hasprop[HasPropertyAttr.MIN_COUNT]:
                self._values[prop_iri].undo()
                raise OldapErrorValue(
                    f'{self.name}: Property {prop_iri} with MIN_COUNT={hasprop[HasPropertyAttr.MIN_COUNT]} has not enough values (n={n}).')

        if hasprop.get(HasPropertyAttr.MAX_COUNT):  # testing for MAX_COUNT conformance
            n = len(self._values[prop_iri])
            if n > hasprop[HasPropertyAttr.MAX_COUNT]:
                self._values[prop_iri].undo()
                raise OldapErrorValue(
                    f'{self.name}: Property {prop_iri} with MAX_COUNT={hasprop[HasPropertyAttr.MIN_COUNT]} has to many values (n={n}).')

        self._changeset[prop_iri] = AttributeChange(None, Action.MODIFY)

    def check_for_permissions(self, permission: AdminPermission) -> (bool, str):
        #
        # First we check if the logged-in user ("actor") has the permission to create a user for
        # the given project!
        #
        actor = self._con.userdata
        sysperms = actor.inProject.get(Iri('oldap:SystemProject', validate=False))
        if sysperms and AdminPermission.ADMIN_OLDAP in sysperms:
            #
            # user has root privileges!
            #
            return True, "OK – IS ROOT"
        else:
            perms = actor.inProject.get(self.project.projectIri)
            if permission in perms:
                return True, "OK",
            else:
                return False, f'Actor does not have {permission} in project "{self.project.projectShortName}".'

    def __get_value(self: Self, prefix: str, fragment: str) -> Xsd | ValueType | None:
        attr = Iri(Xsd_QName(prefix, fragment, validate=False), validate=False)
        tmp = self._values.get(attr, None)
        if tmp is not None and str(attr) in {'oldap:createdBy', 'oldap:creationDate', 'oldap:lastModifiedBy', 'oldap:lastModificationDate'}:
            return next(iter(tmp))
        return tmp

    def __set_value(self: Self, value: ValueType | Xsd | None, prefix: str, fragment: str) -> None:
        prop_iri = Iri(Xsd_QName(prefix, fragment, validate=False), validate=False)
        hasprop = self.properties.get(prop_iri)

        #
        # Validate
        #
        if hasprop.get(HasPropertyAttr.MIN_COUNT):  # testing for MIN_COUNT conformance
            if hasprop[HasPropertyAttr.MIN_COUNT] > 0 and not value:
                raise OldapErrorValue(
                    f'{self.name}: Property {prop_iri} with MIN_COUNT={hasprop[HasPropertyAttr.MIN_COUNT]} is missing')
            elif isinstance(value, (list, tuple, set, ObservableSet)) and len(value) < hasprop[HasPropertyAttr.MIN_COUNT]:
                raise OldapErrorValue(
                    f'{self.name}: Property {prop_iri} with MIN_COUNT={hasprop[HasPropertyAttr.MIN_COUNT]} has not enough values')
        if hasprop.get(HasPropertyAttr.MAX_COUNT):  # testing for MAX_COUNT conformance
            if isinstance(value, (list, tuple, set, ObservableSet)) and len(value) > hasprop[HasPropertyAttr.MAX_COUNT]:
                raise OldapErrorValue(
                    f'{self.name}: Property {prop_iri} with MAX_COUNT={hasprop[HasPropertyAttr.MIN_COUNT]} has to many values (n={len(value)})')
        if value:
            if isinstance(value, LangString):
                self.validate_value(value, hasprop.prop)
            else:
                if isinstance(value, (list, tuple, set, ObservableSet)):
                    for val in self._values[prop_iri]:
                        self.validate_value(val, hasprop.prop)
                else:
                    self.validate_value(value, hasprop.prop)

        if not value:
            self._changeset[prop_iri] = AttributeChange(self._values.get(prop_iri), Action.DELETE)
            del self._values[prop_iri]
        elif isinstance(value, (list, tuple, set, LangString)):  # we may have multiple values...
            if self._values.get(prop_iri):
                self._changeset[prop_iri] = AttributeChange(self._values.get(prop_iri), Action.REPLACE)
            else:
                self._changeset[prop_iri] = AttributeChange(None, Action.CREATE)
            if hasprop.prop.datatype == XsdDatatypes.langString:
                self._values[prop_iri] = LangString(value, notifier=self.notifier, notify_data=prop_iri)
            else:
                self._values[prop_iri] = ObservableSet({
                    convert2datatype(x, hasprop.prop.datatype) for x in value
                }, notifier=self.notifier, notify_data=prop_iri)
        else:
            if self._values.get(prop_iri):
                self._changeset[prop_iri] = AttributeChange(self._values.get(prop_iri), Action.REPLACE)
            else:
                self._changeset[prop_iri] = AttributeChange(None, Action.CREATE)
            try:
                self._values[prop_iri] = ObservableSet({convert2datatype(value, hasprop.prop.datatype)})
            except TypeError:
                self._values[prop_iri] = convert2datatype(value, hasprop.prop.datatype)

    def __del_value(self: Self, prefix: str, fragment: str) -> None:
        prop_iri = Iri(Xsd_QName(prefix, fragment, validate=False), validate=False)
        hasprop = self.properties.get(prop_iri)

        if hasprop.get(HasPropertyAttr.MIN_COUNT):  # testing for MIN_COUNT conformance
            if hasprop[HasPropertyAttr.MIN_COUNT] > 0:
                raise OldapErrorValue(f'{self.name}: Property {prop_iri} with MIN_COUNT={hasprop[HasPropertyAttr.MIN_COUNT]} cannot be deleted.')

        self._changeset[prop_iri] = AttributeChange(self._values.get(prop_iri), Action.DELETE)
        attr = Iri(Xsd_QName(prefix, fragment, validate=False), validate=False)
        del self._values[attr]

    @property
    def iri(self) -> Iri:
        return self._iri

    @property
    def changeset(self) -> dict[Iri, AttributeChange]:
        return self._changeset

    def clear_changeset(self) -> None:
        for item in self._values:
            if hasattr(self._values[item], 'clear_changeset'):
                self._values[item].clear_changeset()
        self._changeset = {}


    def get_data_permission(self, context: Context, permission: DataPermission) -> bool:
        permission_query = context.sparql_context
        permission_query += f'''
        SELECT (COUNT(?permset) as ?numOfPermsets)
        FROM oldap:onto
        FROM shared:onto
        FROM {self._graph}:onto
        FROM NAMED oldap:admin
        FROM NAMED {self._graph}:data
        WHERE {{
            BIND({self._iri.toRdf} as ?iri)
            GRAPH {self._graph}:data {{
                ?iri oldap:grantsPermission ?permset .
            }}
            BIND({self._con.userIri.toRdf} as ?user)
            GRAPH oldap:admin {{
                ?user oldap:hasPermissions ?permset .
                ?permset oldap:givesPermission ?DataPermission .
                ?DataPermission oldap:permissionValue ?permval .
            }}
            FILTER(?permval >= {permission.numeric.toRdf})
        }}'''
        if self._con.in_transaction():
            jsonobj = self._con.transaction_query(permission_query)
        else:
            jsonobj = self._con.query(permission_query)
        res = QueryProcessor(context, jsonobj)
        return res[0]['numOfPermsets'] > 0

    def create(self, indent: int = 0, indent_inc: int = 4) -> str:
        result, message = self.check_for_permissions(AdminPermission.ADMIN_CREATE)
        if not result:
            raise OldapErrorNoPermission(message)

        timestamp = Xsd_dateTimeStamp()
        if self.name == "Thing":
            self._values[Iri('oldap:createdBy', validate=False)] = ObservableSet({self._con.userIri})
            self._values[Iri('oldap:creationDate', validate=False)] = ObservableSet({timestamp})
            self._values[Iri('oldap:lastModifiedBy', validate=False)] = ObservableSet({self._con.userIri})
            self._values[Iri('oldap:lastModificationDate', validate=False)] = ObservableSet({timestamp})

        indent: int = 0
        indent_inc: int = 4

        blank = ''
        context = Context(name=self._con.context_name)
        sparql = context.sparql_context
        sparql += f'{blank:{indent * indent_inc}}INSERT DATA {{'
        sparql += f'\n{blank:{(indent + 1) * indent_inc}}GRAPH {self._graph}:data {{'

        sparql += f'\n{blank:{(indent + 2) * indent_inc}}{self._iri.toRdf} a {self._graph}:{self.name}'

        for prop_iri, values in self._values.items():
            if self.properties.get(prop_iri) and self.properties[prop_iri].prop.datatype == XsdDatatypes.QName:
                qnames = {f'"{x}"^^xsd:QName' for x in values}
                qnames_rdf = ', '.join(qnames)
                sparql += f' ;\n{blank:{(indent + 3) * indent_inc}}{prop_iri.toRdf} {qnames_rdf}'
            else:
                sparql += f' ;\n{blank:{(indent + 3) * indent_inc}}{prop_iri.toRdf} {values.toRdf}'
        sparql += f' .\n{blank:{(indent + 1) * indent_inc}}}}\n'
        sparql += f'{blank:{indent * indent_inc}}}}\n'
        self._con.transaction_start()
        try:
            self._con.transaction_update(sparql)
        except OldapError as e:
            self._con.transaction_abort()
            raise
        self._con.transaction_commit()


    @classmethod
    def read(cls,
             con: IConnection,
             project: Project | Iri | Xsd_NCName | str,
             iri: Iri) -> Self:
        # if not isinstance(project, Project):
        #     project = Project.read(con, project)
        graph = cls.project.projectShortName
        context = Context(name=con.context_name)
        sparql = context.sparql_context
        sparql += f'''
SELECT ?predicate ?value
FROM oldap:onto
FROM shared:onto
FROM {graph}:onto
FROM NAMED oldap:admin
FROM NAMED {graph}:data
WHERE {{
	BIND({iri.toRdf} as ?iri)
    GRAPH {graph}:data {{
        ?iri ?predicate ?value .
        ?iri oldap:grantsPermission ?permset .
    }}
    BIND({con.userIri.toRdf} as ?user)
    GRAPH oldap:admin {{
    	?user oldap:hasPermissions ?permset .
    	?permset oldap:givesPermission ?DataPermission .
    	?DataPermission oldap:permissionValue ?permval .
    }}
    FILTER(?permval >= {DataPermission.DATA_VIEW.numeric.toRdf})
}}'''
        jsonres = con.query(sparql)
        res = QueryProcessor(context, jsonres)
        objtype = None
        kwargs: dict[str, Any] = {}
        for r in res:
            if r['predicate'] == 'rdf:type':
                if r['value'].is_qname:
                    objtype = r['value'].as_qname.fragment
                else:
                    raise OldapErrorInconsistency(f"Expected QName as value, got {r['value']}")
            else:
                if r['predicate'].is_qname:
                    if kwargs.get(r['predicate'].as_qname.fragment):
                        if isinstance(kwargs[r['predicate'].as_qname.fragment], set):
                            kwargs[r['predicate'].as_qname.fragment].add(r['value'])
                        else:
                            kwargs[r['predicate'].as_qname.fragment] = r['value']
                    else:
                        try:
                            kwargs[r['predicate'].as_qname.fragment] = {r['value']}
                        except TypeError:
                            kwargs[r['predicate'].as_qname.fragment] = r['value']
                else:
                    raise OldapErrorInconsistency(f"Expected QName as predicate, got {r['predicate']}")
        if objtype is None:
            raise OldapErrorNotFound(f'Resource with iri <{iri}> not found.')
        if cls.__name__ != objtype:
            raise OldapErrorInconsistency(f'Expected class {cls.__name__}, got {objtype} instead.')
        return cls(iri=iri, **kwargs)

    def update(self, indent: int = 0, indent_inc: int = 4) -> None:
        admin_resources, message = self.check_for_permissions(AdminPermission.ADMIN_RESOURCES)

        context = Context(name=self._con.context_name)
        blank = ''
        timestamp = Xsd_dateTimeStamp()

        sparql_list = []
        required_permission = DataPermission.DATA_EXTEND
        for field, change in self._changeset.items():
            if field == 'oldap:grantsPermission':
                required_permission = DataPermission.DATA_PERMISSIONS
            if change.action == Action.MODIFY:
                continue  # will be processed below!
            if change.action != Action.CREATE:
                if required_permission < DataPermission.DATA_UPDATE:
                    required_permission = DataPermission.DATA_UPDATE
            sparql = f'# Processing field "{field}"\n'
            sparql += f'{blank:{indent * indent_inc}}WITH {self._graph}:data\n'
            if change.action != Action.CREATE:
                sparql += f'{blank:{indent * indent_inc}}DELETE {{\n'
                if isinstance(change.old_value, (ObservableSet, LangString)):
                    for val in change.old_value:
                        sparql += f'{blank:{(indent + 1) * indent_inc}}?res_iri {field.toRdf} {val.toRdf} .\n'
                else:
                    sparql += f'{blank:{(indent + 1) * indent_inc}}?res_iri {field.toRdf} {change.old_value.toRdf} .\n'
                sparql += f'{blank:{indent * indent_inc}}}}\n'
            if change.action != Action.DELETE:
                sparql += f'{blank:{indent * indent_inc}}INSERT {{\n'
                if isinstance(change.old_value, (ObservableSet, LangString)):
                    for val in self._values[field]:
                        sparql += f'{blank:{(indent + 1) * indent_inc}}?res_iri {field.toRdf} {val.toRdf} .\n'
                else:
                    sparql += f'{blank:{(indent + 1) * indent_inc}}?res_iri {field.toRdf} {self._values[field].toRdf} .\n'
                sparql += f'{blank:{indent * indent_inc}}}}\n'
            sparql += f'{blank:{indent * indent_inc}}WHERE {{\n'
            sparql += f'{blank:{(indent + 1) * indent_inc}}BIND({self._iri.toRdf} as ?res_iri)\n'
            if change.action != Action.CREATE:
                if isinstance(change.old_value, (ObservableSet, LangString)):
                    for val in change.old_value:
                        sparql += f'{blank:{(indent + 1) * indent_inc}}?res_iri {field.toRdf} {val.toRdf} .\n'
                else:
                    sparql += f'{blank:{(indent + 1) * indent_inc}}?res_iri {field.toRdf} {change.old_value.toRdf} .\n'
            sparql += f'{blank:{indent * indent_inc}}}}'
            sparql_list.append(sparql)

        for field, change in self._changeset.items():
            if change.action != Action.MODIFY:
                continue  # has been processed above
            if self.properties[field].prop.datatype == XsdDatatypes.langString:
                sparqls = self._values[field].update(graph=f'{self._graph}:data',
                                                     subject=self._iri,
                                                     field=field)
                for lang, lchange in self._values[field].changeset.items():
                    if lchange.action != Action.CREATE:
                        if required_permission < DataPermission.DATA_UPDATE:
                            required_permission = DataPermission.DATA_UPDATE
                sparql_list.extend(sparqls)
            else:
                #
                # first we rectify the datatype of all "new" values added to the set
                #
                newset = {convert2datatype(x, self.properties[field].prop.datatype) for x in self._values[field]}
                self._values[field] = ObservableSet(newset, old_value=self._values[field].old_value, notifier=self.notifier, notify_data=field)
                sparqls = self._values[field].update_sparql(graph=f'{self._graph}:data',
                                                            subject=self._iri,
                                                            field=field)
                sparql_list.extend(sparqls)

        sparql = context.sparql_context
        sparql += f'# Updating resource "{self._iri}"\n'
        sparql += " ;\n".join(sparql_list)

        modtime_update = context.sparql_context
        modtime_update += f'''
WITH {self._graph}:data
DELETE {{
    ?res oldap:lastModificationDate {self.lastModificationDate.toRdf} .
    ?res oldap:lastModifiedBy ?contributor .
}}
INSERT {{
    ?res oldap:lastModificationDate {timestamp.toRdf} .
    ?res oldap:lastModifiedBy {self._con.userIri.toRdf} .
}}
WHERE {{
    BIND({self._iri.toRdf} as ?res)
    ?res oldap:lastModificationDate {self.lastModificationDate.toRdf} .
    ?res oldap:lastModifiedBy ?contributor .
}}
'''

        context = Context(name=self._con.context_name)
        modtime_get = context.sparql_context
        modtime_get += f"""
        SELECT ?modified
        FROM {self._graph}:data
        WHERE {{
            {self._iri.toRdf} oldap:lastModificationDate ?modified
        }}
        """

        self._con.transaction_start()
        #
        # Test permission for Action.REPLACE
        #
        if not admin_resources:
            if not self.get_data_permission(context, required_permission):
                self._con.transaction_abort()
                raise OldapErrorNoPermission(f'No permission to update resource "{self._iri}"')
        try:
            self._con.transaction_update(sparql)
            self._con.transaction_update(modtime_update)
            jsonobj = self._con.transaction_query(modtime_get)
            res = QueryProcessor(context, jsonobj)
            modtime = res[0]['modified']
            if timestamp != modtime:
                raise OldapErrorUpdateFailed(f"Update failed! Timestamp does not match (modtime={modtime}, timestamp={timestamp}).")
        except OldapError:
            self._con.transaction_abort()
            raise
        try:
            self._con.transaction_commit()
        except OldapError:
            self._con.transaction_abort()
            raise
        self._modified = timestamp
        self._contributor = self._con.userIri
        self.clear_changeset()

    def delete(self) -> None:
        admin_resources, message = self.check_for_permissions(AdminPermission.ADMIN_RESOURCES)

        context = Context(name=self._con.context_name)
        inuse = context.sparql_context
        inuse = f'''
        SELECT (COUNT(?res) as ?nres)
        WHERE {{
            ?res ?prop {self._iri.toRdf} .
        }}
        '''

        context = Context(name=self._con.context_name)
        sparql = context.sparql_context
        sparql += f"""
        DELETE WHERE {{
            GRAPH {self._graph}:data {{
                {self._iri.toRdf} ?prop ?val .
            }}
        }} 
        """

        self._con.transaction_start()
        if not admin_resources:
            if not self.get_data_permission(context, DataPermission.DATA_DELETE):
                self._con.transaction_abort()
                raise OldapErrorNoPermission(f'No permission to update resource "{self._iri}"')
        try:
            jsonobj = self._con.transaction_query(inuse)
            res = QueryProcessor(context, jsonobj)
            if res[0]['nres'] > 0:
                raise OldapErrorInUse(f'Resource "{self._iri}" is in use and cannot be deleted.')
        except OldapError:
            self._con.transaction_abort()
            raise
        try:
            self._con.transaction_update(sparql)
        except OldapError:
            self._con.transaction_abort()
            raise
        try:
            self._con.transaction_commit()
        except OldapError:
            self._con.transaction_abort()
            raise


class ResourceInstanceFactory:
    _con: IConnection
    _project: Project
    _datamodel: DataModel

    def __init__(self,
                 con: IConnection,
                 project: Project | Iri | Xsd_NCName | str):
        self._con = con
        if isinstance(project, Project):
            self._project = project
        else:
            self._project = Project.read(self._con, project)

        self._datamodel = DataModel.read(con=self._con, project=self._project)

        #self._oldap_project = Project.read(self._con, "oldap")
        #self._oldap_datamodel = DataModel.read(con=self._con, project=self._oldap_project)

    def createObjectInstance(self, name: Xsd_NCName | str) -> Type:  ## ToDo: Get name automatically from IRI
        classiri = Xsd_QName(self._project.projectShortName, name)
        resclass = self._datamodel.get(classiri)
        if not isinstance(name, Xsd_NCName):
            name = Xsd_NCName(name)
        if resclass is None:
            raise OldapErrorNotFound(f'Given Resource Class "{classiri}" not found.')
        return type(str(name), (ResourceInstance,), {
            '_con': self._con,
            'project': self._project,
            'name': name,
            'factory': self,
            'properties': resclass.properties,
            'superclass': resclass.superclass})








import math
import re
from typing import Self

from pystrict import strict

from oldaplib.src.helpers.oldaperror import OldapErrorValue
from oldaplib.src.helpers.serializer import serializer
from oldaplib.src.xsd.floatingpoint import FloatingPoint
from oldaplib.src.xsd.xsd import Xsd


#@strict
@serializer
class Xsd_double(FloatingPoint):
    """
    Implements the XML Schema [xsd:double](https://www.w3.org/TR/xmlschema11-2/#double) datatype.
    """
    _value: float

    def __init__(self, value: FloatingPoint | float | str, validate: bool = True):
        """
        Constructor for the Xsd_double class.
        :param value: a FloatinPoint instance, a float or a valid string for a float number
        :type value: FloatingPoint | float | str
        :raises OldapErrorValue: if the value cannot be converted to a floating point value
        """
        if isinstance(value, str):
            if validate:
                if not re.match("^([-+]?(\\d+(\\.\\d*)?|\\.\\d+)([eE][-+]?\\d+)?|[Nn]a[Nn]|[-+]?(inf|INF))$", str(value)):
                    raise OldapErrorValue(f'"{value}" is not convertible to a Xsd_float.')
            value = float(value)
        super().__init__(value=value, validate=validate)

    @property
    def toRdf(self) -> str:
        """
        Converts the Xsd_double instance to a RDF string.
        :return: RDF string
        """
        return super()._toRdf('xsd:double')




import re
from datetime import time
from typing import Self

from pystrict import strict

from oldaplib.src.helpers.oldaperror import OldapErrorValue
from oldaplib.src.helpers.serializer import serializer
from oldaplib.src.xsd.xsd import Xsd


#@strict
@serializer
class Xsd_time(Xsd):
    """
    Implements the XML Schema [xsd:time](https://www.w3.org/TR/xmlschema11-2/#time) datatype
    """
    __value: time

    __pattern = r'^([01][0-9]|2[0-3]):([0-5][0-9]):([0-5][0-9]|60)(\.\d+)?(Z|[+-]([01][0-9]|2[0-3]):[0-5][0-9])?$'

    def __init__(self, value: time | Self | str, validate: bool = True):
        """
        Constructor for the Xsd_time class
        :param value: Either a Xsd_time instance, a Python time object, or a valid string representing a time
        :type value: Xsd_time | time | str
        :raises OldapErrorValue: If the value is not a valid time
        """
        if isinstance(value, Xsd_time):
            self.__value = value.__value
        elif isinstance(value, time):
            self.__value = value
        else:
            if validate:
                if re.match(self.__pattern, value) is None:
                    raise OldapErrorValue(f'{value} wrong format for xsd:time.')
            try:
                self.__value = time.fromisoformat(value)
            except ValueError as err:
                raise OldapErrorValue(str(err))

    def __str__(self) -> str:
        """
        String representation of the Xsd_time instance in ISO 8601 format
        :return: String representation of the Xsd_time instance in ISO 8601 format
        :rtype: str
        """
        return self.__value.isoformat()

    def __repr__(self) -> str:
        """
        Python constructor string representation of the Xsd_time instance in ISO 8601 format
        :return: Python constructor string representation of the Xsd_time instance in ISO 8601 format
        :rtype: str
        """
        return f'Xsd_time("{self.__value.isoformat()}")'

    def __eq__(self, other: Self | time | str | None) -> bool:
        """
        Equality check for Xsd_time instance
        :param other: An Xsd_time instance, a Python time object, or a valid string representing a time
        :type other: Xsd_time | time | str | None
        :return: True or False
        :rtype: bool
        :raises OldapErrorValue: If the value is not a valid time
        """
        if other is None:
            return False
        if isinstance(other, Xsd_time):
            return self.__value == other.__value
        if isinstance(other, time):
            return self.__value == other
        if isinstance(other, str):
            if re.match(self.__pattern, other) is None:
                raise OldapErrorValue(f'{other} wrong format for xsd:time.')
            other = time.fromisoformat(other)

    def __hash__(self) -> int:
        """
        Hash representation of the Xsd_time instance in ISO 8601 format
        :return: Hash value
        """
        return hash(str(self))

    def _as_dict(self) -> dict:
        """
        Internal method for serialization to JSON (@serializer decorator)
        :return:
        """
        return {'value': self.__value.isoformat()}

    @property
    def toRdf(self) -> str:
        """
        RDF representation of the Xsd_time instance in ISO 8601 format
        :return: RDF representation of the Xsd_time instance in ISO 8601 format
        :rtype: str
        """
        return f'"{self.__value.isoformat()}"^^xsd:time'

    @property
    def value(self) -> time:
        """
        Time value as Python time object
        :return: Python time value
        """
        return self.__value



import unittest
from copy import deepcopy
from pathlib import Path
from time import sleep

from oldaplib.src.enums.permissionsetattr import PermissionSetAttr
from oldaplib.src.permissionset import PermissionSet
from oldaplib.src.connection import Connection
from oldaplib.src.enums.language import Language
from oldaplib.src.enums.datapermissions import DataPermission
from oldaplib.src.helpers.context import Context
from oldaplib.src.helpers.langstring import LangString
from oldaplib.src.helpers.oldaperror import OldapErrorInconsistency, OldapErrorNotFound, OldapErrorNoPermission, OldapErrorAlreadyExists, OldapErrorImmutable
from oldaplib.src.xsd.iri import Iri
from oldaplib.src.xsd.xsd_qname import Xsd_QName
from oldaplib.src.xsd.xsd_string import Xsd_string


def find_project_root(current_path):
    # Climb up the directory hierarchy and check for a marker file
    path = Path(current_path).absolute()
    while not (path / 'pyproject.toml').exists():
        if path.parent == path:
            # Root of the filesystem, file not found
            raise RuntimeError('Project root not found')
        path = path.parent
    return path


class TestPermissionSet(unittest.TestCase):
    _connection: Connection
    _unpriv: Connection

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        project_root = find_project_root(__file__)

        cls._context = Context(name="DEFAULT")

        cls._connection = Connection(server='http://localhost:7200',
                                     repo="oldap",
                                     userId="rosenth",
                                     credentials="RioGrande",
                                     context_name="DEFAULT")
        cls._unpriv = Connection(server='http://localhost:7200',
                                 repo="oldap",
                                 userId="fornaro",
                                 credentials="RioGrande",
                                 context_name="DEFAULT")

        cls._connection.clear_graph(Xsd_QName('oldap:admin'))
        file = project_root / 'oldaplib' / 'ontologies' / 'admin.trig'
        cls._connection.upload_turtle(file)
        sleep(1)  # upload may take a while...

    @classmethod
    def tearDownClass(cls):
        #cls._connection.clear_graph(Xsd_QName('oldap:admin'))
        #cls._connection.upload_turtle("oldap/ontologies/admin.trig")
        #sleep(1)  # upload may take a while...
        pass

    def test_construct_permissionset(self):
        ps = PermissionSet(con=self._connection,
                           permissionSetId="test1_ps",
                           label=LangString("testPerm@en", "test@Perm@de"),
                           comment=LangString("Testing a PermissionSet@en", "Test eines PermissionSet@Perm@de"),
                           givesPermission=DataPermission.DATA_UPDATE,
                           definedByProject=Iri('oldap:SystemProject'))
        self.assertEqual(ps.givesPermission, DataPermission.DATA_UPDATE)
        self.assertEqual(ps.label, LangString("testPerm@en", "test@Perm@de"))
        self.assertEqual(ps.comment, LangString("Testing a PermissionSet@en", "Test eines PermissionSet@Perm@de"))
        self.assertEqual(ps.definedByProject, Iri('oldap:SystemProject'))

    def test_deepcopy_permissionset(self):
        ps = PermissionSet(con=self._connection,
                           permissionSetId="test1_ps",
                           label=LangString("testPerm@en", "test@Perm@de"),
                           comment=LangString("Testing a PermissionSet@en", "Test eines PermissionSet@Perm@de"),
                           givesPermission=DataPermission.DATA_UPDATE,
                           definedByProject=Iri('oldap:SystemProject'))
        ps1 = deepcopy(ps)
        self.assertEqual(ps.permissionSetId, ps1.permissionSetId)
        self.assertFalse(ps.permissionSetId is ps1.permissionSetId)

        self.assertEqual(ps.label, ps1.label)
        self.assertFalse(ps.label is ps1.label)

        self.assertEqual(ps.comment, ps1.comment)
        self.assertFalse(ps.comment is ps1.comment)

        self.assertEqual(ps.givesPermission, ps1.givesPermission)

        self.assertEqual(ps.definedByProject, ps1.definedByProject)
        self.assertFalse(ps.definedByProject is ps1.definedByProject)


    def test_create_permissionset(self):
        ps = PermissionSet(con=self._connection,
                           permissionSetId="test3_ps",
                           label=LangString("\";SELECT * { password ?p ?o . }@en", "test@Perm@de"),
                           comment=LangString("Testing a PermissionSet@en", "Test eines PermissionSet@Perm@de"),
                           givesPermission=DataPermission.DATA_UPDATE,
                           definedByProject=Iri('oldap:SystemProject'))
        ps.create()
        self.assertIsNotNone(ps.created)
        self.assertIsNotNone(ps.creator)
        self.assertIsNotNone(ps.modified)
        self.assertIsNotNone(ps.contributor)
        del ps
        ps = PermissionSet.read(self._connection, "test3_ps", Iri('oldap:SystemProject'), ignore_cache=True)
        self.assertEqual(ps.givesPermission, DataPermission.DATA_UPDATE)
        self.assertEqual(ps.label, LangString("\";SELECT * { password ?p ?o . }@en", "test@Perm@de"))
        self.assertEqual(ps.comment, LangString("Testing a PermissionSet@en", "Test eines PermissionSet@Perm@de"))
        self.assertEqual(ps.definedByProject, Iri('oldap:SystemProject'))

    def test_create_permset_with_shortname(self):
        ps = PermissionSet(con=self._connection,
                           permissionSetId="test4_ps",
                           label=LangString("test4@en", "test4@Perm@de"),
                           comment=LangString("Testing a PermissionSet 4@en", "Test eines PermissionSet 4@Perm@de"),
                           givesPermission=DataPermission.DATA_UPDATE,
                           definedByProject='britnet')
        ps.create()
        ps = PermissionSet.read(self._connection, "test4_ps", Iri('http://www.salsah.org/version/2.0/SwissBritNet'), ignore_cache=True)
        self.assertEqual(ps.givesPermission, DataPermission.DATA_UPDATE)
        self.assertEqual(ps.label, LangString("test4@en", "test4@Perm@de"))

    def test_create_prmission_set_without_label(self):
        ps = PermissionSet(con=self._connection,
                           permissionSetId="test5_ps",
                           label=LangString(),
                           comment=LangString("Testing a PermissionSet 4@en", "Test eines PermissionSet 4@Perm@de"),
                           givesPermission=DataPermission.DATA_UPDATE,
                           definedByProject='britnet')
        ps.create()
        self.assertEqual(ps.givesPermission, DataPermission.DATA_UPDATE)

    # @unittest.skip('Work in progress')
    def test_read_permission_A(self):
        ps = PermissionSet.read(self._connection, 'GenericView', 'oldap:SystemProject', ignore_cache=True)
        self.assertEqual(ps.givesPermission, DataPermission.DATA_VIEW)  # add assertion here
        self.assertEqual(ps.label, LangString("GenericView@en", "GenericView@de", "GenericView@fr", "GenericView@it"))
        self.assertEqual(ps.definedByProject, Iri('oldap:SystemProject'))

    def test_read_permission_B(self):
        ps = PermissionSet.read(self._connection, 'HyperHamletMember', 'oldap:HyperHamlet', ignore_cache=True)
        self.assertEqual(ps.givesPermission, DataPermission.DATA_UPDATE)  # add assertion here
        self.assertEqual(ps.label, LangString("HyHaUpdate@en", "HyHaUpdate@de", "HyHaUpdate@fr", "HyHaUpdate@it"))
        self.assertEqual(ps.definedByProject, Iri('oldap:HyperHamlet'))

    def test_read_permission_with_cache(self):
        ps = PermissionSet.read(self._connection, 'HyperHamletMember', 'oldap:HyperHamlet', ignore_cache=True)
        ps2 = PermissionSet.read(self._connection, 'HyperHamletMember', 'oldap:HyperHamlet')
        self.assertEqual(ps2.givesPermission, DataPermission.DATA_UPDATE)
        self.assertEqual(ps.givesPermission, ps2.givesPermission)

        self.assertEqual(ps2.label, LangString("HyHaUpdate@en", "HyHaUpdate@de", "HyHaUpdate@fr", "HyHaUpdate@it"))
        self.assertEqual(ps.label, ps2.label)
        self.assertFalse(ps.label is ps2.label)

        self.assertEqual(ps2.definedByProject, Iri('oldap:HyperHamlet'))
        self.assertEqual(ps.definedByProject, ps2.definedByProject)
        self.assertFalse(ps.definedByProject is ps2.definedByProject)


    def test_create_permission(self):
        ps = PermissionSet(con=self._connection,
                           permissionSetId="testPerm",
                           label=LangString("testPerm@en", "test@Perm@de"),
                           comment=LangString("Testing a PermissionSet@en", "Test eines PermissionSet@Perm@de"),
                           givesPermission=DataPermission.DATA_UPDATE,
                           definedByProject=Iri('oldap:SystemProject'))
        ps.create()
        self.assertIsNotNone(ps.created)
        self.assertIsNotNone(ps.creator)
        self.assertIsNotNone(ps.modified)
        self.assertIsNotNone(ps.contributor)
        del ps
        ps = PermissionSet.read(self._connection, "testPerm", 'oldap:SystemProject', ignore_cache=True)
        self.assertEqual(ps.givesPermission, DataPermission.DATA_UPDATE)
        self.assertEqual(ps.label, LangString("testPerm@en", "test@Perm@de"))
        self.assertEqual(ps.comment, LangString("Testing a PermissionSet@en", "Test eines PermissionSet@Perm@de"))
        self.assertEqual(ps.definedByProject, Iri('oldap:SystemProject'))

    def test_create_permission_strange_label(self):
        ps = PermissionSet(con=self._connection,
                           permissionSetId="gagaPerm",
                           label=LangString("\";SELECT * { password ?p ?o . }@en", "test@Perm@de"),
                           comment=LangString("Testing a PermissionSet@en", "Test eines PermissionSet@Perm@de"),
                           givesPermission=DataPermission.DATA_UPDATE,
                           definedByProject=Iri('oldap:SystemProject'))
        ps.create()
        self.assertIsNotNone(ps.created)
        self.assertIsNotNone(ps.creator)
        self.assertIsNotNone(ps.modified)
        self.assertIsNotNone(ps.contributor)
        del ps
        ps = PermissionSet.read(self._connection, "gagaPerm", 'oldap:SystemProject', ignore_cache=True)
        self.assertEqual(ps.givesPermission, DataPermission.DATA_UPDATE)
        self.assertEqual(ps.label, LangString("\";SELECT * { password ?p ?o . }@en", "test@Perm@de"))
        self.assertEqual(ps.comment, LangString("Testing a PermissionSet@en", "Test eines PermissionSet@Perm@de"))
        self.assertEqual(ps.definedByProject, Iri('oldap:SystemProject'))

    def test_create_permission_no_label(self):
        ps = PermissionSet(con=self._connection,
                           permissionSetId="gagaPerm2",
                           givesPermission=DataPermission.DATA_UPDATE,
                           definedByProject=Iri('oldap:SystemProject'))
        ps.create()
        del ps
        ps = PermissionSet.read(self._connection, "gagaPerm2", 'oldap:SystemProject', ignore_cache=True)
        self.assertEqual(ps.givesPermission, DataPermission.DATA_UPDATE)
        self.assertIsNone(ps.label)
        self.assertIsNone(ps.comment)
        self.assertEqual(ps.definedByProject, Iri('oldap:SystemProject'))

    def test_create_permission_duplicate(self):
        ps = PermissionSet(con=self._connection,
                           permissionSetId="testPerm2",
                           label=LangString("testPerm@en", "test@Perm@de"),
                           comment=LangString("Testing a PermissionSet@en", "Test eines PermissionSet@Perm@de"),
                           givesPermission=DataPermission.DATA_UPDATE,
                           definedByProject=Iri('oldap:SystemProject'))
        ps.create()
        ps = PermissionSet(con=self._connection,
                           permissionSetId="testPerm2",
                           label=LangString("testPerm33@en", "test@Perm33@de"),
                           comment=LangString("Testing a PermissionSet33@en", "Test eines PermissionSet33@Perm@de"),
                           givesPermission=DataPermission.DATA_UPDATE,
                           definedByProject=Iri('oldap:SystemProject'))
        with self.assertRaises(OldapErrorAlreadyExists) as ex:
            ps.create()


    def test_create_permission_unauthorized(self):
        ps = PermissionSet(con=self._unpriv,
                           permissionSetId="testPermUnauth",
                           label=LangString("testPermUnauth@en", "test@PermUnauth@de"),
                           comment=LangString("Testing a PermissionSet@en", "Test eines PermissionSet@Perm@de"),
                           givesPermission=DataPermission.DATA_UPDATE,
                           definedByProject=Iri('oldap:SystemProject'))
        with self.assertRaises(OldapErrorNoPermission):
            ps.create()

        def test_create_permission_unauthorized(self):
            ps = PermissionSet(con=self._unpriv,
                               label=LangString("testPermUnauth2@en", "test@PermUnauth2@de"),
                               comment=LangString("Testing a PermissionSet@en", "Test eines PermissionSet@Perm@de"),
                               givesPermission=DataPermission.DATA_UPDATE,
                               definedByProject=Iri('oldap:HyperHamlet'))
            with self.assertRaises(OldapErrorNoPermission):
                ps.create()

    # TODO: More testing!!!
    def test_search_permission_sets(self):
        iris = PermissionSet.search(self._connection, label="GenericView")
        self.assertEqual(len(iris), 2)
        self.assertTrue(Iri('oldap:GenericView') in iris)

        iris = PermissionSet.search(self._connection, label=Xsd_string("GenericView@de"))
        self.assertEqual(len(iris), 2)
        self.assertTrue(Iri('oldap:GenericView') in iris)

        iris = PermissionSet.search(self._connection, definedByProject=Iri("oldap:HyperHamlet"))
        self.assertEqual(len(iris), 1)
        self.assertEqual(Iri('hyha:HyperHamletMember'), iris[0])

        iris = PermissionSet.search(self._connection, givesPermission=DataPermission.DATA_RESTRICTED)
        #self.assertEqual(len(iris), 1)
        self.assertEqual([Iri('oldap:GenericRestricted')], iris)

    def test_update_permission_set(self):
        ps = PermissionSet(con=self._connection,
                           permissionSetId="testUpdatePerm",
                           label=LangString("testUpdatePerm@en", "testVerändernPerm@de"),
                           comment=LangString("Testing update of PermissionSet@en", "Test einer Veränderung eines PermissionSet@Perm@de"),
                           givesPermission=DataPermission.DATA_UPDATE,
                           definedByProject=Iri('oldap:SystemProject'))
        ps.create()
        psId = ps.permissionSetId
        del ps
        ps = PermissionSet.read(self._connection, psId, Iri('oldap:SystemProject'), ignore_cache=True)
        ps.label[Language.FR] = "testeModificationPerm"
        ps.givesPermission = DataPermission.DATA_VIEW
        ps.update()
        ps = PermissionSet.read(self._connection, psId, Iri('oldap:SystemProject'))
        self.assertEqual(ps.givesPermission, DataPermission.DATA_VIEW)
        self.assertEqual(ps.label, LangString("testUpdatePerm@en", "testVerändernPerm@de", "testeModificationPerm@fr"))
        del ps.comment
        ps.update()
        self.assertIsNone(ps.comment)
        self.assertIsNone(ps.get(PermissionSetAttr.COMMENT))

        ps = PermissionSet.read(self._connection, psId, Iri('oldap:SystemProject'), ignore_cache=True)
        ps.comment = LangString("gagagaga@en")
        with self.assertRaises(OldapErrorImmutable):
            ps[PermissionSetAttr.DEFINED_BY_PROJECT] = Iri('oldap:HyperHamlet')

        ps = PermissionSet.read(self._unpriv, psId, Iri('oldap:SystemProject'), ignore_cache=True)
        ps.comment = LangString("gagagaga@fr")
        with self.assertRaises(OldapErrorNoPermission):
            ps.update()

    def test_update_permissionset_B(self):
        ps = PermissionSet(con=self._connection,
                           permissionSetId="testUpdatePermB",
                           label=LangString("testUpdatePerm@en", "testVerändernPerm@de"),
                           comment=LangString("Testing update of PermissionSet@en", "Test einer Veränderung eines PermissionSet@Perm@de"),
                           givesPermission=DataPermission.DATA_UPDATE,
                           definedByProject=Iri('oldap:SystemProject'))
        ps.create()
        psId = ps.permissionSetId
        del ps
        ps = PermissionSet.read(self._connection, psId, Iri('oldap:SystemProject'), ignore_cache=True)

        ps.label[Language.IT] = "TEST_ADD_DEL"
        del ps.label[Language.EN]
        ps.update()

    def test_delete_permission_set(self):
        ps = PermissionSet(con=self._connection,
                           permissionSetId="testDeletePerm",
                           label=LangString("testDeletePerm@en", "testDeletePerm@de"),
                           comment=LangString("Testing deleting a PermissionSet@en", "Test einer Löschung eines PermissionSet@Perm@de"),
                           givesPermission=DataPermission.DATA_UPDATE,
                           definedByProject=Iri('oldap:HyperHamlet'))
        ps.create()
        del ps
        ps = PermissionSet.read(self._connection, "testDeletePerm", Iri('oldap:HyperHamlet'), ignore_cache=True)
        ps.delete()

        with self.assertRaises(OldapErrorNotFound) as ex:
            project = ps.read(self._connection, "testDeletePerm", Iri('oldap:HyperHamlet'), ignore_cache=True)


if __name__ == '__main__':
    unittest.main()



from typing import Any, Self, Dict

from pystrict import strict

from oldaplib.src.helpers.oldaperror import OldapErrorValue
from oldaplib.src.helpers.serializer import serializer
from oldaplib.src.xsd.xsd import Xsd


#@strict
@serializer
class Xsd_boolean(Xsd):
    """
    Implements the XML Schema [xsd:boolean](https://www.w3.org/TR/xmlschema11-2/#boolean) datatype
    """
    __value: bool

    def __init__(self, value: Any, validate: bool = True):
        """
        Constructor of Xsd_boolean
        :param value: Any value that can be interpreted as boolean
        :type value: Any
        :param validate: Boolean value that determines whether or not the value should be validated (not used!)
        :raises OldapErrorValue: If the value is not a boolean
        """
        if isinstance(value, str):
            if value.lower() in ('yes', 'true', 't', 'y', '1'):
                self.__value = True
            elif value.lower() in ('no', 'false', 'f', 'n', '0'):
                self.__value = False
            else:
                raise OldapErrorValue('No valid string for boolean value.')
        else:
            self.__value = bool(value)

    def __str__(self) -> str:
        """
        String representation of Xsd_boolean
        :return: string
        :rtype: str
        """
        return str(self.__value).lower()

    def __repr__(self) -> str:
        """
        String representation of Xsd_boolean as constructor
        :return: string
        :rtype: str
        """
        return f"Xsd_boolean('{str(self.__value).lower()}')"

    def __bool__(self) -> bool:
        """
        Boolean representation of Xsd_boolean
        :return: bool
        :rtype: bool
        """
        return self.__value

    def __eq__(self, other: Any | None) -> bool:
        """
        Equality check for Xsd_boolean
        :param other: Value to compare with
        :type other: Any
        :return: True or False
        :rtype: bool
        """
        if other is None:
            return False
        if not isinstance(other, Xsd_boolean):
            other = Xsd_boolean(other)
        return self.__value == other.__value

    @property
    def toRdf(self) -> str:
        """
        String representation of Xsd_boolean for RDF representation
        :return: string for RDF
        :rtype: str
        """
        return f'"{str(self.__value).lower()}"^^xsd:boolean'

    @classmethod
    def fromRdf(cls, rdf: str) -> Self:
        return cls(rdf)

    def _as_dict(self) -> dict[str, str]:
        """
        Used internally for JSON serialization using @serialisation decorator
        :return: dict
        :rtype: dict[str, str]
        """
        return {'value': str(self.__value)}

    @property
    def value(self) -> bool:
        """
        Boolean representation of Xsd_boolean
        :return: bool
        """
        return self.__value



from oldaplib.src.propertyclass import PropertyClass
from oldaplib.src.xsd.floatingpoint import FloatingPoint
from oldaplib.src.xsd.iri import Iri
from oldaplib.src.xsd.xsd_float import Xsd_float
from oldaplib.src.xsd.xsd_integer import Xsd_integer


class PropRef:
    """
    This class represents a numeric value that eiter can be a subclass of Xsd_integer (and it's subclasses) or
    a subclass of FloatingPoint (and it's subclasses). FLoatingPoint is the superclass of all XML Schema datatypes
    with a floating point content (e.g. Xsd_float, Xsd_decimal etc.)
    """
    def __new__(cls, value: PropertyClass | Iri | str):
        if isinstance(value, (PropertyClass)):
            return value
        elif isinstance(value, (Iri, str)):
            return Iri(value)


if __name__ == '__main__':
    a = PropRef("sh:gaga")
    print(type(a).__name__)



from typing import Self

from pystrict import strict

from oldaplib.src.helpers.oldaperror import OldapErrorValue
from oldaplib.src.helpers.serializer import serializer
from oldaplib.src.xsd.xsd import Xsd


#@strict
@serializer
class Xsd_integer(Xsd):
    """
    Base class for XSD Schema integer classes, implements directly the XSD Schema
    [xsd:integer](https://www.w3.org/TR/xmlschema11-2/#integer) datatype
    """
    _value: int

    def __init__(self, value: Xsd | int | str, validate: bool = True):
        """
        Constructor for Xsd_integer
        :param value: Value to convert to Xsd_integer
        :type value: Xsd | int | str
        :raises OldapErrorValue: Value cannot be converted to Xsd_integer
        """
        if isinstance(value, Xsd_integer):
            self._value = value._value
        elif isinstance(value, int):
            self._value = value
        else:
            try:
                self._value = int(value)
            except ValueError as err:
                raise OldapErrorValue(str(err))

    def __str__(self) -> str:
        """
        String representation of Xsd_integer
        :return: string
        """
        return str(self._value)

    def __repr__(self) -> str:
        """
        Constrctor string representation of Xsd_integer in constructor form
        :return: string
        """
        return f'{type(self).__name__}({str(self._value)})'

    def __hash__(self) -> int:
        """
        Hash value of Xsd_integer
        :return: Hash value
        """
        return hash(self._value)

    def __int__(self) -> int:
        """
        Converts Xsd_integer to integer
        :return: integer value
        """
        return self._value

    def __eq__(self, other: Self | int | None) -> bool:
        """
        Equality check for Xsd_integer
        :param other: Value to compare instance with
        :type other: Xsd | int | None
        :return: True or False
        :raises OldapErrorValue: Value cannot be converted to Xsd_integer
        """
        if other is None:
            return False
        if isinstance(other, Xsd_integer):
            return self._value == other._value
        elif isinstance(other, int):
            return self._value == other
        else:
            raise OldapErrorValue(f'Comparison of with {type(other)} not possible')

    def __ne__(self, other: Self | int) -> bool:
        """
        Inequality check for Xsd_integer
        :param other: Value to compare with
        :type other: Xsd | int | None
        :return: True of False
        :raises OldapErrorValue: Value cannot be converted to Xsd_integer
        """
        if other is None:
            return True
        if isinstance(other, Xsd_integer):
            return self._value != other._value
        elif isinstance(other, int):
            return self._value != other
        else:
            raise OldapErrorValue(f'Comparison of with {type(other)} not possible')

    def __gt__(self, other: Self | int) -> bool:
        """
        Compare for greater than value
        :param other: Value to compare with
        :return: True or False
        :raises OldapErrorValue: Value cannot be converted to Xsd_integer
        """
        if isinstance(other, Xsd_integer):
            return self._value > other._value
        elif isinstance(other, int):
            return self._value > other
        else:
            raise OldapErrorValue(f'Comparison of with {type(other)} not possible')

    def __ge__(self, other: Self | int) -> bool:
        """
        Compare for greater or equal than value
        :param other: Value to compare with
        :return: True or False
        :raises OldapErrorValue: Value cannot be converted to Xsd_integer
        """
        if isinstance(other, Xsd_integer):
            return self._value >= other._value
        elif isinstance(other, int):
            return self._value >= other
        else:
            raise OldapErrorValue(f'Comparison of with {type(other)} not possible')

    def __lt__(self, other: Self | int) -> bool:
        """
        Compare for less than value
        :param other: Value to compare with
        :return: True or False
        :raises OldapErrorValue: Value cannot be converted to Xsd_integer
        """
        if isinstance(other, Xsd_integer):
            return self._value < other._value
        elif isinstance(other, int):
            return self._value < other
        else:
            raise OldapErrorValue(f'Comparison of with {type(other)} not possible')

    def __le__(self, other: Self | int) -> bool:
        """
        Compare for less or equal than value
        :param other: Value to compare with
        :return: True or False
        :raises OldapErrorValue: Value cannot be converted to Xsd_integer
        """
        if isinstance(other, Xsd_integer):
            return self._value <= other._value
        elif isinstance(other, int):
            return self._value <= other
        else:
            raise OldapErrorValue(f'Comparison of with {type(other)} not possible')

    def _as_dict(self) -> dict[str, int]:
        """
        Internal method for JSON serialization based on @serializer decorator
        :return: dict
        """
        return {'value': self._value}

    @property
    def toRdf(self) -> str:
        """
        Converts Xsd_integer to RDF string
        :return: RDF string representation of Xsd_integer
        """
        xsddummy, name = type(self).__name__.split('_')
        return f'"{str(self._value)}"^^xsd:{name}'

    @property
    def value(self) -> int:
        """
        Converts Xsd_integer to integer
        :return: integer
        """
        return self._value




from enum import unique, Enum
from typing import Dict

from oldaplib.src.helpers.serializer import serializer


@unique
@serializer
class Action(Enum):
    """
    # Action

    An Enumeration of the Actions that are supported on PropertyClass and ResourceClass attributes/restrictions

    - `Action.CREATE` = 'create'
    - `Action.MODIFY` = 'modify'
    - `Action.REPLACE` = 'replace'
    - `Action.DELETE` = 'delete'
    """
    CREATE = 'create'  # a new value has been added
    MODIFY = 'modify'  # a complex value (LangString, PropertyRestriction) has been modified
    REPLACE = 'replace'  # an existing value has been replaced by a new value
    DELETE = 'delete'  # an existing value has been deleted

    def _as_dict(self) -> Dict[str, str]:
        return {__class__: self.__class__.__name__, 'value': self.value}



import json
from typing import Iterable

from pystrict import strict

from oldaplib.src.dtypes.rdfset import RdfSet
from oldaplib.src.helpers.Notify import Notify
from oldaplib.src.helpers.oldaperror import OldapErrorType, OldapErrorInconsistency
from oldaplib.src.helpers.serializer import serializer
from oldaplib.src.xsd.xsd import Xsd
from oldaplib.src.xsd.xsd_string import Xsd_string


@serializer
class XsdSet(RdfSet[Xsd]):
    """
    The XsdSet is a subclass of [RdfSet](/python_docstrings/rdfset) that restricts the set values to
    subclasses of the Xsd class, that is Xsd compatible data types.
    """

    def __init__(self, *args: Iterable[Xsd] | Xsd, value: Iterable[Xsd] | Xsd | None = None):
        #
        # Here we first do some fancy type checking, before we call the superclass' __init__() method
        #
        if len(args) == 0:
            if value is None:
                pass
            else:
                if isinstance(value, Iterable):
                    for v in value:
                        if not isinstance(v, Xsd):
                            raise OldapErrorType(f'Iterable contains element that is not an instance of "Xsd", but "{type(v).__name__}".')
                else:
                    if not isinstance(value, Xsd):
                        raise OldapErrorType(f'Value is not an instance of "Xsd", but "{type(value).__name__}".')
        elif len(args) == 1:
            if isinstance(args[0], Iterable):
                for v in args[0]:
                    if not isinstance(v, Xsd):
                        raise OldapErrorType(f'Iterable contains element that is not an instance of "Xsd", but "{type(v).__name__}".')
            else:
                if not isinstance(args[0], Xsd):
                    raise OldapErrorType(f'Value is not an instance of "Xsd", but "{type(args[0]).__name__}".')
        else:
            for arg in args:
                if not isinstance(args[0], Xsd):
                    raise OldapErrorType(f'Value is not an instance of "Xsd", but "{type(args[0]).__name__}".')
        super().__init__(*args, value=value)

    def add(self, val: Xsd) -> None:
        self.notify()

        if isinstance(val, Xsd) and not type(val) is Xsd:
            self._data.add(val)
        else:
            if not self._data:
                raise OldapErrorInconsistency(f"It's not possible to add {val} to an empty set.")
            item_type = type(next(iter(self._data)))
            self._data.add(item_type(val))


if __name__ == "__main__":
    gaga = XsdSet(None)


from datetime import timedelta
from typing import Self

import isodate
from isodate import ISO8601Error
from pystrict import strict

from oldaplib.src.helpers.oldaperror import OldapErrorValue
from oldaplib.src.helpers.serializer import serializer
from oldaplib.src.xsd.xsd import Xsd


#@strict
@serializer
class Xsd_duration(Xsd):
    """
    Implements the XML Schema [xsd:duration](https://www.w3.org/TR/xmlschema11-2/#duration) datatype
    """
    __value: timedelta

    def __init__(self, value: timedelta | Self | str, validate: bool = True):
        """
        Constructor of the Xsd_duration class.
        :param value: a Xsd_duration instance, or a timedelta instance, or a string for a timedelta
        in ISO8601 format.
        :type value: Self | timedelta | str
        :raises OldapErrorValue: if the value is not a valid timedelta
        """
        if isinstance(value, Xsd_duration):
            self.__value = value.__value
        elif isinstance(value, timedelta):
            self.__value = value
        else:
            try:
                self.__value = isodate.parse_duration(value)
            except ISO8601Error as err:
                raise OldapErrorValue(str(err))

    def __str__(self) -> str:
        """
        String representation of the Xsd_duration instance.
        :return: String representation
        """
        return isodate.duration_isoformat(self.__value)

    def __repr__(self) -> str:
        """
        Constructor string representation of the Xsd_duration instance.
        :return: Constructor string
        """
        return f'"{isodate.duration_isoformat(self.__value)}"^^xsd:duration'

    def __eq__(self, other: Self | timedelta | str | None) -> bool:
        """
        Compares two Xsd_duration instances for equality.
        :param other: Xsd_duration instance, a timedelta instance, or a string for a timedelta
        :type other: Self | timedelta | str | None
        :return: True or False
        :raise OldapErrorValue: if the value is not a valid timedelta
        """
        if other is None:
            return False
        if isinstance(other, Xsd_duration):
            return self.__value == other.__value
        if isinstance(other, timedelta):
            return self.__value == other
        else:
            try:
                other = isodate.parse_duration(str(other))
            except ISO8601Error as err:
                raise OldapErrorValue(str(err))
            return self.__value == other

    def __hash__(self) -> int:
        """
        Hash the Xsd_duration instance.
        :return: Hast value
        """
        return hash(self.__value)

    @property
    def toRdf(self) -> str:
        """
        Converts the Xsd_duration instance to a RDF string.
        :return: RDF string
        """
        return f'"{isodate.duration_isoformat(self.__value)}"^^xsd:duration'

    def _as_dict(self) -> dict:
        """
        Internal method to convert for JSON serialization. (@serializer decorater)
        :return:
        """
        return {'value': isodate.duration_isoformat(self.__value)}

    @property
    def value(self) -> timedelta:
        """
        Converts the Xsd_duration instance to a timedelta instance.
        :return: timedelta instance
        """
        return self.__value



from enum import unique

from oldaplib.src.dtypes.namespaceiri import NamespaceIRI
from oldaplib.src.enums.attributeclass import AttributeClass
from oldaplib.src.helpers.langstring import LangString
from oldaplib.src.xsd.iri import Iri
from oldaplib.src.xsd.xsd_date import Xsd_date
from oldaplib.src.xsd.xsd_ncname import Xsd_NCName


@unique
class ProjectAttr(AttributeClass):
    """
    This enum class represents the fields used in the project model
    """
    # order: (QName, mandatory, immutable, datatype)
    PROJECT_IRI = ('oldap:projectIri', False, True, Iri)  # virtual property, repents the RDF subject
    PROJECT_SHORTNAME = ('oldap:projectShortName', True, True, Xsd_NCName)
    LABEL = ('rdfs:label', False, False, LangString)
    COMMENT = ('rdfs:comment', False, False, LangString)
    NAMESPACE_IRI = ('oldap:namespaceIri', True, True, NamespaceIRI)
    PROJECT_START = ('oldap:projectStart', False, False, Xsd_date)
    PROJECT_END = ('oldap:projectEnd', False, False, Xsd_date)
